/*
 * Copyright (C) 2011-2012 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//`define SUPPORT_PG		// ALU_WIDTH == 4 only
//`define SUPPORT_SIO_COMATIBLE

//
// Module Name: alu281
//   ver 001
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                           ALU_WIDTH:   4         8
// Number of Slice Flip Flops:            4         8
// Number of 4 input LUTs:               79       115
// Number of occupied Slices:            40        60
// Total Number of 4 input LUTs:         79       116
// Number of bonded IOBs:                22        30

// Post-PAR Static Timing Report
// Maximum frequency(MHz):           144.238    97.466
//
//        original 74281
//    A1  1            24  VCC  
//    A2  2            23  A0 
//   RS1  3            22  CP
//   RS0  4            21  SIO0
//    RC  5            20  AS0
//  SIO3  6            19  AS1
//    A3  7            18  AS2
//    Cn  8            17  M  
//    ~G  9            16  F0
//  Cn+4  10           15  F1
//    ~P  11           14  F2
//   GND  12           13  F3
//
// XO2-256        modified 74281      XO2-256
//  27        A1  1            24  VCC  
//  28        A2  2            23  A0    25
//  29       RS1  3            22  CP    23
//  30       RS0  4            21  SIO0  21
//  32        RC  5            20  AS0   20
//   1      SIO3  6            19  AS1   17
//   4        A3  7            18  AS2   16
//   5        Cn  8          x 17  N.C.
//   8      N.C.  9  x         16  F0    14
//   9      Cn+4  10           15  F1    13
//  10         M  11 x         14  F2    12
//           GND  12           13  F3    11
	   

`define SUPPORT_SIO_COMATIBLE

module alu281 # (
	parameter ALU_WIDTH = 4
) (
		 input   C_IN       // Carry Input
		, input  [ALU_WIDTH-1:0] A     // DATA IN
		, output [ALU_WIDTH-1:0] F   // DATA OUT
		, output  C_OUT      // carry-look-ahead
`ifdef SUPPORT_PG
		, output  P          // carry-propagate 
		, output  G          // carry-generate
`endif
		, input   M          // Mode Control
		, input   [2:0] AS   // Function Select
		, input   [1:0] RS   // Reg. Select
		, input   RC         // Reg. Control
		, input   CP         // Clock 
`ifdef SUPPORT_SIO_COMATIBLE
		, inout   SIO0        // RI/LO
		, inout   SIO3        // LI/RO
`else
		, input   SI0        // RI/LO
		, output  SO0        // RI/LO
		, input   SI3        // LI/RO
		, output  SO3        // LI/RO
`endif
		);
	wire [ALU_WIDTH-1:0] B;

	alu_74281 # ( .ALU_WIDTH(ALU_WIDTH) ) ALU (
		  .C_IN(C_IN)     // Carry Input
		, .A(A[ALU_WIDTH-1:0])      // DATA IN
		, .B(B[ALU_WIDTH-1:0])      // DATA IN
		, .F(F[ALU_WIDTH-1:0])      // DATA OUT
		, .C_OUT(C_OUT)   // carry-look-ahead
`ifdef SUPPORT_PG
		, .P(P)           // carry-propagate 
		, .G(G)           // carry-generate
`endif
		, .M(M)           // Mode Control
		, .AS(AS[2:0])    // Function Select
		);
	acc_74281 # ( .ALU_WIDTH(ALU_WIDTH)) ACC (
		  .B(B[ALU_WIDTH-1:0])      // DATA OUT
		, .F(F[ALU_WIDTH-1:0])      // DATA IN

		, .RS(RS[1:0])    // Reg. Select
		, .RC(RC)         // Reg. Control
		, .CP(CP)         // Clock 
		, .SI0(SI0)       // RI/LO
		, .SO0(SO0)       // RI/LO
		, .SI3(SI3)       // LI/RO
		, .SO3(SO3)       // LI/RO
		);

`ifdef SUPPORT_SIO_COMATIBLE
	assign SIO0 = (RS == 2) ? SO0 : 1'bz ;
	assign SIO3 = (RS == 1) ? SO3 : 1'bz ;

	assign SI0 = SIO0;
	assign SI3 = SIO3;
`endif

endmodule

module alu_74281 # (
	parameter ALU_WIDTH = 4
) (
		  input   C_IN       // Carry Input
		, input  [ALU_WIDTH-1:0] A    // DATA IN
		, input  [ALU_WIDTH-1:0] B    // DATA IN
		, output [ALU_WIDTH-1:0] F    // DATA OUT
		, output  C_OUT      // carry-look-ahead
`ifdef SUPPORT_PG
		, output  P          // carry-propagate 
		, output  G          // carry-generate
`endif

		, input   M          // Mode Control
		, input  [2:0] AS   // Function Select
		);

`ifdef SUPPORT_PG
    wire [3:0] int_p = 
		  ~M & (AS == 1) ?  ~A |  B
		: ~M & (AS == 2) ?   A | ~B
		: ~M & (AS == 3) ?   A |  B
		: ~M & (AS == 4) ?    B
		: ~M & (AS == 5) ?   ~B
		: ~M & (AS == 6) ?    A
		: ~M & (AS == 7) ?   ~A
		: M & (AS == 1)  ?  ~A |  B
		: M & (AS == 2)  ?   A |  B
		: M & (AS == 3)  ?  ~A |  B
		: 0;

    wire [3:0] int_g = 
		  ~M & (AS == 1) ?  ~A &  B
		: ~M & (AS == 2) ?   A & ~B
		: ~M & (AS == 3) ?   A &  B
		: M & (AS == 1)  ?  ~A &  B
		: M & (AS == 2)  ?   A &  B
		: M & (AS == 3)  ?  ~A &  B
		: M & (AS == 4)  ?   A &  B
		: M & (AS == 5)  ?  ~A & ~B
		: M & (AS == 6)  ?   A &  B
		: M & (AS == 7)  ?  ~A & ~B
		: 1'b0;

    assign P = ~M & (AS != 0) | M & ((AS == 1) | (AS == 2) |(AS == 3))
		    ?  ~( int_p[3] int_p[2] int_p[1] int_p[0]) : 0 ;

    assign G = ((AS == 1) | (AS == 2) |(AS == 3))
		    ? ~( int_g[3] | int_p[3] & int_g[2] 
			 | int_p[3] & int_p[2] & int_g[1] 
			 | int_p[3] & int_p[2] & int_p[1] & int_g[0] )
	     : M & ((AS == 4) | (AS == 5) | (AS == 6) | (AS == 7))
		   ? ~( int_g[3] & int_g[2] & int_g[1] & int_g[0] )
	     : 1'b1;
`endif
    wire [ALU_WIDTH-2:0] minus1 = -1;

`ifdef ZZZ
    wire [ALU_WIDTH:0] v_out =
		  ~M & (AS == 0) ? { 1'b0, minus1 , C_IN }
		: ~M & (AS == 1) ?  B - A - C_IN
		: ~M & (AS == 2) ?  A - B - C_IN
		: ~M & (AS == 3) ?  A + B + C_IN
		: ~M & (AS == 4) ? ~B + C_IN
		: ~M & (AS == 5) ?  B + C_IN
		: ~M & (AS == 6) ?  A + C_IN
		: ~M & (AS == 7) ? ~A + C_IN
		: M & (AS == 1) ? { 1'b0,  (A ^ B) }
		: M & (AS == 2) ? { 1'b0, ~(A ^ B) }
		: M & (AS == 3) ? { 1'b0,  (A ^ B) }
		: M & (AS == 4) ? { 1'b0,  (A & B) }
		: M & (AS == 5) ? { 1'b0, ~(A | B) }
		: M & (AS == 6) ? { 1'b0, ~(A & B) }
		: M & (AS == 7) ? { 1'b0,  (A | B) }
		: 0;
`else
    wire [ALU_WIDTH-1:0] AoB = ((AS[1]) ? A : B);
    wire [ALU_WIDTH:0] plus_out = ((AS == 4) | (AS == 7)) ? { 1'b0, ~AoB }
				: ((AS == 5) | (AS == 6)) ? { 1'b0,  AoB }
				: ( A + B );
    wire [ALU_WIDTH:0] minus_out = (AS[0]) ? (B - A) : (A - B) ;
    wire [ALU_WIDTH-1:0] logic_out =
		  ( (AS == 4) |(AS == 6) ) ? (A & B)
		: ( (AS == 5) |(AS == 7) ) ? (A | B)
		: (A ^ B) ;
    wire [ALU_WIDTH:0] v_out =
		  ~M & (AS == 0) ? { 1'b0, minus1 , C_IN }
		: ~M & ( (AS == 1) |(AS == 2) )
			? minus_out - C_IN
		: ~M ? plus_out + C_IN
		: M & (AS == 0) ? 0
		: M & ((AS == 2) | (AS == 5) | (AS == 6) )
			? { 1'b0, ~logic_out}
		: { 1'b0, logic_out};
`endif
    assign F     = v_out[ALU_WIDTH-1:0];
    assign C_OUT = v_out[ALU_WIDTH];

endmodule

module acc_74281 # (
	parameter ALU_WIDTH = 4
) (
		  output [ALU_WIDTH-1:0] B    // DATA OUT
		, input  [ALU_WIDTH-1:0] F    // DATA IN

		, input  [1:0] RS   // Reg. Select
		, input   RC         // Reg. Control
		, input   CP         // Clock 
		, input   SI0        // RI/LO
		, output  SO0        // RI/LO
		, input   SI3        // LI/RO
		, output  SO3        // LI/RO
		);

    reg [ALU_WIDTH-1:0] acc;

    assign B = acc; 
    assign SO0 =  F[0];
    assign SO3 = RC ? F[ALU_WIDTH-2] : F[ALU_WIDTH-1];

    always @(posedge CP)
    begin
	acc  <=   (RS == 0) ? F
		: (RS == 1) ? { ( RC ? acc[ALU_WIDTH-1] : F[ALU_WIDTH-2] )
				, F[ALU_WIDTH-3:0] , SI0 }
		: (RS == 2) ? { ( RC ? acc[ALU_WIDTH-1] : SI3 )
				, ( RC ? SI3 : F[ALU_WIDTH-1] )
				, F[ALU_WIDTH-2:1] }
		: acc;
    end

    initial
    begin
	acc  <= 0;
    end

endmodule
